@@ -53,6 +53,8 @@ struct _GeditDocumentsPanelPrivate
 	GtkWidget          *treeview;
 	GtkTreeModel       *model;
 
+	guint               refresh_idle_id;
+
 	guint               adding_tab : 1;
 	guint               is_reodering : 1;
 	guint               setting_active_notebook : 1;
@@ -161,85 +163,84 @@ get_iter_from_tab (GeditDocumentsPanel *panel,
 
 	gedit_debug (DEBUG_PANEL);
 
-	g_assert (notebook != NULL || tab != NULL);
+	g_assert (notebook != NULL);
+	/* tab may be NULL if just the notebook is specified */
 	g_assert (tab_iter != NULL);
-	g_assert ((tab == NULL && tab_iter == NULL) || (tab != NULL && tab_iter != NULL));
 
-	success = FALSE;
 	search_notebook = (gedit_multi_notebook_get_n_notebooks (panel->priv->mnb) > 1);
 
-	/* We never ever ask for a notebook or tab if it does not already exist in the tree */
-	g_assert (gtk_tree_model_get_iter_first (panel->priv->model, &parent));
+	if (!gtk_tree_model_get_iter_first (panel->priv->model, &parent))
+		return FALSE;
 
+	success = FALSE;
 	do
 	{
+		GtkTreeIter iter;
+
 		if (search_notebook)
 		{
 			GeditNotebook *current_notebook;
-			gboolean is_cur;
 
 			gtk_tree_model_get (panel->priv->model,
 					    &parent,
 					    NOTEBOOK_COLUMN, &current_notebook,
 					    -1);
 
-			is_cur = (current_notebook == notebook);
-
-			g_object_unref (current_notebook);
-
-			if (is_cur)
+			if (current_notebook != NULL)
 			{
-				success = TRUE;
+				gboolean is_cur;
+				is_cur = (current_notebook == notebook);
 
-				if (tab == NULL)
+				g_object_unref (current_notebook);
+
+				if (is_cur)
 				{
-					break;
+					/* notebook found */
+					search_notebook = FALSE;
+					gtk_tree_model_iter_children (panel->priv->model, &iter, &parent);
 				}
 			}
 		}
-		else if (tab == NULL)
+		else
 		{
-			success = TRUE;
-
-			break;
+			iter = parent;
 		}
 
-		if (success || !search_notebook)
+		if (!search_notebook)
 		{
-			GeditTab *current_tab;
-			GtkTreeIter iter;
+			if (tab == NULL)
+			{
+				success = TRUE;
+				break;
+			}
 
-			if (search_notebook)
-			{
-				g_assert (gtk_tree_model_iter_children (panel->priv->model,
-									&iter, &parent));
-			}
-			else
-			{
-				iter = parent;
-			}
+			g_assert (gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model),
+								&iter));
 
 			do
 			{
-				gboolean is_cur;
+				GeditTab *current_tab;
 
 				gtk_tree_model_get (panel->priv->model,
 						    &iter,
 						    TAB_COLUMN, &current_tab,
 						    -1);
-				g_assert (current_tab != NULL);
 
-				is_cur = (current_tab == tab);
-
-				g_object_unref (current_tab);
-
-				if (is_cur)
+				if (current_tab != NULL)
 				{
-					*tab_iter = iter;
-					success = TRUE;
+					gboolean is_cur;
 
-					/* break 2; */
-					goto out;
+					is_cur = (current_tab == tab);
+					g_object_unref (current_tab);
+
+					if (is_cur)
+					{
+						*tab_iter = iter;
+						success = TRUE;
+
+						/* break 2; */
+						goto out;
+					}
 				}
 			} while (gtk_tree_model_iter_next (panel->priv->model, &iter));
 
@@ -250,15 +251,6 @@ get_iter_from_tab (GeditDocumentsPanel *panel,
 
 out:
 
-	/* In the current code if we are not successful then something is wrong */
-	g_assert (success);
-
-	if (tab_iter != NULL)
-	{
-		g_assert (gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model),
-							tab_iter));
-	}
-
 	return success;
 }
 
@@ -297,8 +289,8 @@ select_active_tab (GeditDocumentsPanel *panel)
 	{
 		GtkTreeIter iter;
 
-		get_iter_from_tab (panel, notebook, tab, &iter);
-		select_iter (panel, &iter);
+		if (get_iter_from_tab (panel, notebook, tab, &iter))
+			select_iter (panel, &iter);
 	}
 }
 
@@ -317,10 +309,8 @@ multi_notebook_tab_switched (GeditMultiNotebook  *mnb,
 	{
 		GtkTreeIter iter;
 
-		get_iter_from_tab (panel, new_notebook, new_tab, &iter);
-
-		if (gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model),
-						  &iter))
+		if (get_iter_from_tab (panel, new_notebook, new_tab, &iter) &&
+		    gtk_tree_store_iter_is_valid (GTK_TREE_STORE (panel->priv->model), &iter))
 		{
 			select_iter (panel, &iter);
 		}
@@ -417,8 +407,8 @@ refresh_notebook_foreach (GeditNotebook       *notebook,
 	}
 }
 
-static void
-refresh_list (GeditDocumentsPanel *panel)
+static gboolean
+refresh_list_idle (GeditDocumentsPanel *panel)
 {
 	gedit_debug (DEBUG_PANEL);
 
@@ -433,6 +423,22 @@ refresh_list (GeditDocumentsPanel *panel)
 	gtk_tree_view_expand_all (GTK_TREE_VIEW (panel->priv->treeview));
 
 	select_active_tab (panel);
+
+	panel->priv->refresh_idle_id = 0;
+
+	return FALSE;
+}
+
+static void
+refresh_list (GeditDocumentsPanel *panel)
+{
+	/* refresh in an idle so that when adding/removing many tabs
+	 * the model is repopulated just once */
+	if (panel->priv->refresh_idle_id == 0)
+	{
+		panel->priv->refresh_idle_id = gdk_threads_add_idle ((GSourceFunc) refresh_list_idle,
+		                                                     panel);
+	}
 }
 
 static void
@@ -453,29 +459,32 @@ sync_name_and_icon (GeditTab            *tab,
 		    GParamSpec          *pspec,
 		    GeditDocumentsPanel *panel)
 {
-	GdkPixbuf *pixbuf;
-	gchar *name;
 	GtkTreeIter iter;
 
 	gedit_debug (DEBUG_PANEL);
 
-	get_iter_from_tab (panel,
-			   gedit_multi_notebook_get_active_notebook (panel->priv->mnb),
-			   tab, &iter);
-
-	name = tab_get_name (tab);
-	pixbuf = _gedit_tab_get_icon (tab);
-
-	gtk_tree_store_set (GTK_TREE_STORE (panel->priv->model),
-			    &iter,
-			    PIXBUF_COLUMN, pixbuf,
-			    NAME_COLUMN, name,
-			    -1);
-
-	g_free (name);
-	if (pixbuf != NULL)
+	if (get_iter_from_tab (panel,
+			       gedit_multi_notebook_get_active_notebook (panel->priv->mnb),
+			       tab,
+			       &iter))
 	{
-		g_object_unref (pixbuf);
+		gchar *name;
+		GdkPixbuf *pixbuf;
+
+		name = tab_get_name (tab);
+		pixbuf = _gedit_tab_get_icon (tab);
+
+		gtk_tree_store_set (GTK_TREE_STORE (panel->priv->model),
+				    &iter,
+				    PIXBUF_COLUMN, pixbuf,
+				    NAME_COLUMN, name,
+				    -1);
+
+		g_free (name);
+		if (pixbuf != NULL)
+		{
+			g_object_unref (pixbuf);
+		}
 	}
 }
 
@@ -682,6 +691,12 @@ gedit_documents_panel_dispose (GObject *object)
 
 	gedit_debug (DEBUG_PANEL);
 
+	if (panel->priv->refresh_idle_id != 0)
+	{
+		g_source_remove (panel->priv->refresh_idle_id);
+		panel->priv->refresh_idle_id = 0;
+	}
+
 	if (panel->priv->window != NULL)
 	{
 		g_object_unref (panel->priv->window);
@@ -1015,17 +1030,17 @@ close_button_clicked (GtkCellRenderer     *cell,
                       const gchar         *path,
                       GeditDocumentsPanel *panel)
 {
-        GtkTreeIter iter;
-        GeditTab *tab;
-        GeditNotebook *notebook;
+	GtkTreeIter iter;
+	GeditTab *tab;
+	GeditNotebook *notebook;
 
-        if (!gtk_tree_model_get_iter_from_string (panel->priv->model,
-                                                  &iter, path))
-        {
-                return;
-        }
+	if (!gtk_tree_model_get_iter_from_string (panel->priv->model,
+	                                          &iter, path))
+	{
+	        return;
+	}
 
-        gtk_tree_model_get (panel->priv->model,
+	gtk_tree_model_get (panel->priv->model,
 		            &iter,
 		            NOTEBOOK_COLUMN, &notebook,
 		            TAB_COLUMN, &tab,
@@ -1037,7 +1052,8 @@ close_button_clicked (GtkCellRenderer     *cell,
 	}
 	else
 	{
-		gedit_notebook_remove_tab (notebook, tab);
+		gtk_container_remove (GTK_CONTAINER (notebook),
+		                      GTK_WIDGET (tab));
 		g_object_unref (tab);
 	}
 
